# -*- coding: utf-8 -*-

"""Main module."""

import numpy as np
import numba
from scipy.integrate import  quad, simps

class NumericZ:
    """A class that can store the information from the numerical calculation of Z
    Then useing the __call__  methode to compute the Z function.
    """

    def __init__(self, F_function, N=64):
        self.F = F_function
        self.N = N
        self.delt = 1

        self.L = np.sqrt(self.N / np.sqrt(2))
        self.compute_a_real()
        self.compute_a_complex()

        self.imag_tol = 100000

    def setimagetol(self, value):
        self.imag_tol = value

    def compute_a_real(self):
        """Compute the a_n coefficients for the real axis"""
        N = self.N
        n = np.arange(-N, N)
        v = self.L * np.tan(np.pi * (n + 1 / 2) / (2 * N))
        FF = self.F(v)
        a = np.fft.fft(np.fft.fftshift(FF * (self.L - 1.j * v)))
        a = np.exp(- 1.j * n * np.pi / (2 * N)) * np.fft.fftshift(a)
        a = np.flipud(1.j * np.sign(n + 1 / 2) * a) / (2 * N)

        self.areal = a

    def compute_a_complex(self):
        """Compute the a_n coefficients for the complex domain"""
        N = self.N  # Two times more points for the complex domain

        n = np.arange(-N, N)
        v = self.L * np.tan(np.pi * (n) / (2 * N))
        FF = self.F(v) * (self.L ** 2 + v ** 2)

        a = np.fft.fft(np.fft.fftshift(FF)) / (2 * N)
        a0 = a[0]
        a = np.flipud(a[1:N])

        self.aimag0 = a0
        self.aimag = a

    def hilb1(self, zf):
        """Hilbert transfor on the real axis"""
        t = (self.L + 1.j * zf) / (self.L - 1.j * zf)
        #  Real axis
        h = polyval(self.areal, t) / (t ** self.N * (self.L - 1.j * zf))
        return h

    def hilb2(self, zf):

        t = np.array((self.L + 1.j * zf) / (self.L - 1.j * zf))
        if t.size > 0:
            p = polyval(self.aimag, t)
            h2 = 1.j * (2 * p / (self.L - 1.j * zf) ** 2 + (self.aimag0 / self.L) / (self.L - 1.j * zf))
        else:
            h2 = np.array([])
        return h2

    def hilb(self, w):
        """hilb"""
        zvalues = np.zeros(w.shape, dtype=np.complex)

        #  Real axis
        realindex = np.nonzero(np.imag(w) == 0)

        if len(realindex) > 0:
            wreal = w[realindex]
            h = self.hilb1(wreal)
            FFreal = self.F(wreal)
            zvalues[realindex] = h + 1.j*FFreal

        # Complex domain
        compindex = np.nonzero(np.imag(w) != 0)
        
        if len(compindex) > 0:
            wcomp = w[compindex]
            zf = w[compindex]
            
            negativimag = np.nonzero(zf.imag < 0)
            positivimag = np.nonzero(zf.imag != 0)
            
            #zf[negativimag] = np.conj(zf[negativimag])

            h2 = self.hilb2(zf)
            #FFcomp = self.F(wcomp[negativimag])

            try:
                #zvalues[negativimag] = np.conj(h2[negativimag]) #+ self.delt*2.j*FFcomp
                zvalues[positivimag] = h2[positivimag]
            except TypeError:
                print(zvalues)
                print(h2)
                print(negativimag)


        return np.pi * zvalues
    
    def __call__(self, w):

        if not isinstance(w, np.ndarray):
            w = np.array([w])
        try:
            z = np.zeros(len(w), dtype=complex)
        except TypeError:
            print(w, type(w))
            print(w.shape)

        #indexreal = np.nonzero(np.abs(w.imag) <= self.imag_tol)
        # indexreal = np.arange(w.shape[0])

        #if len(indexreal) > 0:
        z = self.hilb(w)

        indexcomp = np.argwhere(np.abs(w.imag) > self.imag_tol)

        # indexcomp = []
        def integrand(x, zi):
            return self.F(x) / (x - zi)

        if len(indexcomp) > 0:
            try:
                for i, zi in enumerate(w[indexcomp]):
                    tmp_v = complex_quadrature(integrand, -10, 10, zi=zi)
                    # tmp_v = complex_quadrature(integrand, -10,10, args=(zi,), epsabs=1.49e-04, epsrel=1.49e-04,
                    #                                    full_output = 1)
                    z[indexcomp[i]] = tmp_v

            except IndexError:
                print(i, indexcomp[i], zi)

        return z


@numba.jit(nopython=True)
def polyval(p, x):
    """Faster (?) than numpy.polyval
    """

    y = np.zeros_like(x)
    for i in range(len(p)):
        y = y * x + p[i]
    return y




def complex_quadrature_quad(func, a, b, **kwargs):
    def real_func(x, *args):
        return func(x, *args).real

    def imag_func(x, *args):
        return func(x, *args).imag

    real_integral = quad(real_func, a, b, **kwargs)
    imag_integral = quad(imag_func, a, b, **kwargs)
    return real_integral[0] + 1j * imag_integral[0]


def complex_quadrature(func, a, b, **kwargs):
    v_axis = np.linspace(-10, 10, 512)
    f_points = func(v_axis, **kwargs)
    integral = simps(f_points, x=v_axis)

    return integral





import numpy as np



def generalized_Maxw_f(v, a,b,c,d):
    """general distribution function used for the fits"""
    return a * np.exp( - (v**2)**(c/2)/d)


def maxw_f(v, vt):
    return np.exp(- (v/vt)**2)/np.sqrt(np.pi)/vt





from numerical_z.plasma import PlasmaParameters
from numerical_z.numerical_z import NumericZ

from numerical_z.evdfs import maxw_f
from plasmapy.mathematics import plasma_dispersion_func

import numpy as np
import scipy as sc

"Normalized permittivity"


def eps_IAW(omg, kx, ky, kz, prt=PlasmaParameters(),
            Ze=plasma_dispersion_func,
            Zi=plasma_dispersion_func):
    def Zep(x):
        return -2 * (1 + x * Ze(x))

    def Zip(x):
        return -2 * (1 + x * Zi(x))

    kp = np.sqrt(ky ** 2 + kz ** 2)
    eBeta = kp ** 2 * prt.ope ** 2 / prt.oce ** 2
    k2 = kx ** 2 + kp ** 2

    if k2 == 0:
        raise RuntimeError("The wave vector is Zero !!")

    iChi = (omg) / (ky * prt.vti)
    eChi = (omg - ky * prt.vde) / (ky * prt.vte)

    iEps = 1 / (ky**2 * prt.vti ** 2) * Zip(iChi)
    eEps = prt.ope ** 2 / (ky**2 * prt.vte ** 2) * Zep(eChi)
    #eEps = -1 / (ky**2 )
    #iEps =  (1 + 3/2*ky**2/omg**2 * prt.vti**2)/omg**2
    #iEps =  1/omg**2


    return 1 - iEps - eEps

def analytic_IAW(k, ti=0):
    return np.sqrt(1/(1 + 1/(k**2))*(1 + 3*ti*(1+k**2)))


def analytic_IAW_simple(k, ti=0):
    return np.sqrt(1/(1 + 1/(k**2)))





from numerical_z.plasma import PlasmaParameters
from numerical_z.numerical_z import NumericZ

from numerical_z.evdfs import maxw_f
from plasmapy.mathematics import plasma_dispersion_func

import numpy as np


"Normalized permittivity"

def epsLangmuir_Maxw(omg, kx, ky, kz):
    eChi = omg / (np.sqrt(2) * ky)

    Z_echi = plasma_dispersion_func(eChi)
    Z_echip = -2 * (1 + eChi * Z_echi)

    eEps = 1 / (2 * ky ** 2) * Z_echip

    return (1 - eEps)



def epsLangmuirGene(omg, kx, ky, kz, Z=plasma_dispersion_func):
    eChi = omg / (np.sqrt(2) * ky)

    Z_echi = Z(eChi)
    Z_echip = -2 * (1 + eChi * Z_echi)

    eEps = 1 / (2 * ky ** 2) * Z_echip

    return (1 - eEps)

def langmuir_analytic(k, prt):
    return np.sqrt( prt.ope**2 + 3 * k * prt.vte)





from numerical_z.plasma import PlasmaParameters
from numerical_z.numerical_z import NumericZ

from numerical_z.evdfs import maxw_f
from plasmapy.mathematics import plasma_dispersion_func

import numpy as np
import scipy as sc

"Normalized permittivity"


def eps_ECDI(omg, kx, ky, kz, prt=PlasmaParameters(), Zlocal=plasma_dispersion_func, Nmax = 5):
    def Zp(x):
        return -2 * (1 + x * Zlocal(x))

    kp = np.sqrt(ky ** 2 + kz ** 2)
    eBeta = kp ** 2 * prt.ope ** 2 / prt.oce ** 2
    k2 = kx ** 2 + kp ** 2

    if k2 == 0:
        raise RuntimeError("The wave vector is Zero !!")

    iChi = (omg - kz * prt.vdi) / (np.sqrt(k2) * prt.vti)
    eChi = (omg - ky * prt.vde) / (kx * prt.vte)
    bessSum = np.array([sc.special.iv(n, eBeta) * Zlocal(eChi - n * prt.oce / (kx * prt.vte)) for n in range(-Nmax, Nmax)])

    iEps = 1 / (k2 * prt.vti ** 2) * Zp(iChi)
    eEps = 1 / k2 * (1. + np.exp(-eBeta) * eChi * sum(bessSum))

    return 1 - iEps + eEps



# -*- coding: utf-8 -*-

"""Top-level package for  Numerical_Z."""

__author__ = """Antoine Tavant"""
__email__ = 'antoinetavant@hotmail.fr'
__version__ = '0.1.0'


from numerical_z.numerical_z import  NumericZ
from numerical_z.plasma import PlasmaParameters
from numerical_z.langmuir import  epsLangmuir_Maxw, epsLangmuirGene



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

requirements = ["numpy", "numba", "scipy", "plasmapy",  ]

setup_requirements = [ ]

test_requirements = [ ]

setup(
    author="Antoine Tavant",
    author_email='antoinetavant@hotmail.fr',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="Compute the plasma equation Z from any EVDF",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description='\n\n',
    include_package_data=True,
    keywords='numerical_z',
    name='numerical_z',
    packages=find_packages(include=['numerical_z']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://hephaistos.lpp.polytechnique.fr/rhodecode/GIT_REPOSITORIES/LPP/Users/Tavant/DevCalcul/Numerical_Z',
    version='0.1.1',
    zip_safe=False,
)



import numpy as np

class PlasmaParameters:
    """Parameters needed for the Dispersion Relation """

    def __init__(self, ne=1e17, te=50, ti=0.1, bf=0.02, ef=4e4):
        """sets the parameters of the Plasma"""
        self.qe = 1.6e-19  # "Electric charge (C)"
        self.c = 3e8  # "Speed of light (m/s)"
        self.me = 9.1e-31  # "Electron mass (Kg)"
        self.eps0 = 8.85e-12  # "Vacuum permittivity(F/m)"
        self.mi = 2.18e-25  # "Ion mass (Kg)"

        "Parameters of the system"
        self.ne = ne  # "Plasma density (m^-3)"
        self.te = te  # "Electron temperature (eV)"
        self.ti = ti  # "Ion temperature (eV)"
        self.bf = bf  # "Magnetic field (T)"
        self.ef = ef  # "Electric field (V/m)"

        self.opi = np.sqrt(self.ne * self.qe ** 2 / (self.eps0 * self.mi))  # "Ion plasma frequency"
        self.ld = np.sqrt(self.eps0 * self.te / (self.ne * self.qe))  # "Debye length"
        self.cs = self.ld * self.opi  # "Sound speed"
        "Derived quantities (normalized)"

        self.ope = np.sqrt(self.ne * self.qe ** 2 / (self.eps0 * self.me)) / self.opi  # "Ion plasma frequency"
        self.oce = self.qe * self.bf / (self.me * self.opi)  # "Electron cyclotron frquency"
        self.vte = np.sqrt(2 * self.qe * self.te / self.me) / self.cs  # "Electron thermal velocity"
        self.vde = self.ef / (self.bf * self.cs)  # "Electron drift velocity"
        # vde=ef/bf
        # vde = 4.5*vte

        self.vti = np.sqrt(2 * self.qe * self.ti / self.mi) / self.cs  # "Ion thermal velocity"
        self.rL = self.vte / self.oce  # "Larmor radius"
        self.eps = 0.01
        self.vtb = 0.5 * self.vte
        self.vdi = 10000 / self.cs
